class PagingInformation:
    """
    Data structure to store paging information. The first page is page 0
    """

    def __init__(self, page=0, page_size=25):
        """
        Creates a new object with the provided parameters.
        """
        self.page = page
        self.page_size = page_size



import time
from datetime import datetime
from abc import ABC, abstractmethod
from pbu.logger import Logger
from pbu.constant_listing import ConstantListing


class JobStatus(ConstantListing):
    CREATED = "CREATED"
    QUEUED = "QUEUED"
    PENDING = "PENDING"
    RUNNING = "RUNNING"
    FINISHED = "FINISHED"
    COMPLETED = "COMPLETED"
    FAILED = "FAILED"
    STALLED = "STALLED"
    SUCCESS = "SUCCESS"
    ERROR = "ERROR"
    PAUSED = "PAUSED"
    RESUMED = "RESUMED"


class BasicMonitor(ABC):
    """
    Abstract base class (ABC) for all monitors containing base functionality for lifecycle and meta information.
    """

    def __init__(self, monitor_id, wait_time, run_interval=False, custom_logger=None, ping_interval=60):
        """
        Super constructor invoked from implementing sub-classes of this class
        providing interfaces start, track_success and track_error
        """
        # set meta flags
        self.active = False
        self.finished = False
        self.started = False

        # store parameters
        self.monitor_id = monitor_id
        self.wait_time = wait_time
        self.ping_interval = ping_interval
        self.run_interval = run_interval

        # init logger
        if custom_logger is not None:
            self.logger = custom_logger
        else:
            self.logger = Logger(self.__class__.__name__)
        # handle ping interval issues
        if wait_time < ping_interval:
            self.logger.info(f"WARNING, monitor wait time {ping_interval} is longer than {wait_time} - overriding")
            self.ping_interval = wait_time
        # runtime variables
        self.next_execution = 0

    @abstractmethod
    def running(self):
        """
        Abstract main method of each inverter. `running` needs to be implemented by every non-abstract sub-class of this
        class.
        """
        pass

    def to_json(self):
        """
        Creates a dictionary with some base meta attributes about this monitor. Sub classes should invoke this method
        when serialising to json and then add any additional parameters as necessary to the result of this methods
        invoke.
        :return: a dictionary with basic information, such as site_id and overall active status
        """
        return {
            "active": self.active,
            "started": self.started,
            "finished": self.finished,
            "id": self.monitor_id,
            "waitTime": self.wait_time,
            "runInterval": self.run_interval,
        }

    def wait(self, exec_duration=0):
        """
        Executes a wait for a certain amount of time depending on the configuration of the monitor. Additionally, it
        takes into account the execution time of the last loop, if run_interval is True.
        :param exec_duration: the execution time of the last execution loop (only relevant for run_interval=True)
        """
        now = time.time()
        # set next execution
        if self.run_interval:
            gap_next = max(1, self.wait_time - exec_duration)
            if gap_next < self.ping_interval:
                # special handling if next execution is sooner than ping interval
                time.sleep(gap_next)
                return  # exit after wait
            self.next_execution = now + gap_next
        else:
            self.next_execution = now + self.wait_time

        # check every ping interval if we're close to next execution
        while time.time() + self.ping_interval < self.next_execution:
            time.sleep(self.ping_interval)

        # wait remaining time
        remaining = round(self.next_execution - time.time())
        if remaining > 0:
            time.sleep(remaining)
        if remaining < 0:
            self.logger.info(f"Overdue execution by {-round(remaining)}s")

    def wait_till_midnight(self):
        """
        This method will just wait until the next midnight event. If the monitor is started at 6:30pm, this method will
        simply wait for 5h and 30min.
        """
        # localise current date
        local_date = datetime.now()
        # extract time
        local_time = local_date.time()
        # compute wait time until midnight
        wait_time = ((23 - local_time.hour) * 60 * 60) + ((59 - local_time.minute) * 60) + (60 - local_time.second)
        # wait till midnight
        self.logger.info("Waiting {} seconds for monitor {} until midnight".format(wait_time, self.monitor_id))
        time.sleep(wait_time)

    def start(self):
        """
        Starts the monitor, updating a flag and triggering `running()`
        """
        if self.active:
            # already started
            return
        # start monitor
        self.active = True
        try:
            if self.started:
                # only restart if the previous thread has finished or after error
                self.logger.info("Restarting monitor for {}".format(self.monitor_id))
                self.running()
            else:
                # first time start
                self.started = True
                self.logger.info("Starting monitor for {}".format(self.monitor_id))
                self.running()
            self.finished = True
        except BaseException as ex:
            self.logger.exception("Exception during monitor execution for monitor {}: {}".format(self.monitor_id,
                                                                                                 str(ex)))
            # is currently not active due to error
            self.active = False
            # wait for one execution loop to avoid error spamming
            time.sleep(self.wait_time)
            self.start()

    def stop(self):
        """
        Sets the active flag to false, which will terminate the `running()` loop.
        """
        # set a flag, let the monitor handle this
        self.logger.info("Stopping {} for {}".format(self.__class__.__name__, self.monitor_id))
        self.active = False



from pytz import timezone, utc, BaseTzInfo
from datetime import datetime, date, time
from typing import Union, Optional

DATETIME_FORMAT = "%Y-%m-%d %H:%M:%S"
DATE_FORMAT = "%Y-%m-%d"


def to_timezone(localized_datetime: datetime, target_timezone: Union[BaseTzInfo, str]) -> datetime:
    """
    Translates a localized or unlocalized datetime into a date of the given timezone. The resulting datetime object will
    be offset b the number of hours difference between the input timezone and the target timezone.
    :param localized_datetime: a datetime object with or without a tz info set
    :param target_timezone: the timezone to translate the input to, this can be provided as string (name)or pytz
    timezone object.
    :return: a datetime object in the provided timezone with a different hour value than the input parameter.
    """
    if isinstance(target_timezone, str):
        tz = timezone(target_timezone)
        return to_timezone(localized_datetime, tz)

    return datetime.fromtimestamp(localized_datetime.timestamp(), tz=target_timezone)


def to_utc(localized_datetime: datetime) -> datetime:
    """
    Translates a localized datetime of some timezone into an UTC date. The resulting datetime object will be offset by
    the number of hours difference between the input timezone and UTC.
    :param localized_datetime: a datetime object with or without a tz info set.
    :return: a datetime object in UTC with different hour value than the input parameter.
    """
    return to_timezone(localized_datetime, utc)


def combine_date_time(date_obj: date, time_obj: time, opt_timezone: Optional[Union[BaseTzInfo, str]]) -> datetime:
    """
    Combines a date and time object into a datetime object and optionally sets the timezone as well. If the timezone is
    provided, the datetime object will NOT be shifted/translated from your local timezone, but merely it will set the
    tzinfo value for the resulting datetime object.
    :param date_obj: the date component for the resulting datetime object
    :param time_obj: the time component for the resulting datetime object
    :param opt_timezone: (optional) timezone specification for the resulting datetime object. This can be either
    provided as string (name of the timezone) or pytz timezone object.
    :return: a datetime object as combination of the provided date and time elements with an optional timezone info.
    """
    base_dt = datetime(date_obj.year, date_obj.month, date_obj.day, time_obj.hour, time_obj.minute, time_obj.second,
                       time_obj.microsecond)
    if opt_timezone is None:
        return base_dt

    tz = timezone(opt_timezone) if isinstance(opt_timezone, str) else opt_timezone
    return tz.localize(base_dt)


def set_timezone(unlocalized_datetime: datetime, target_timezone: Union[BaseTzInfo, str]) -> datetime:
    """
    Simply sets the timezone part of the datetime for a given datetime object. The provided datetime can already have a
    tzinfo set, in which case it will simply be replaced.
    :param unlocalized_datetime: a datetime object with or without a tz info
    :param target_timezone: a timezone either provided as string (name of timezone) or pytz timezone object.
    :return: a datetime object with the same hour/minute values as the provided input, but the timezone set according to
    the provided parameter.
    """
    if unlocalized_datetime.tzinfo is not None:
        # remove current tz info and call this function again
        return set_timezone(unlocalized_datetime.replace(tzinfo=None), target_timezone)

    if isinstance(target_timezone, str):
        return set_timezone(unlocalized_datetime, timezone(target_timezone))

    return target_timezone.localize(unlocalized_datetime)



import math
import pandas as pd
from copy import copy
from pbu.json_wrapper import JSON
from datetime import datetime, timedelta

DEFAULT_DATE_FORMAT = "%Y-%m-%d %H:%M:%S"


class TimeSeries:
    """
    Helper class to manage time series with multiple data points. It offers the ability to align dates of different time
    series, force a fixed resolution and interpolate missing values, add another time series to an existing one or
    simply fill columns with defaults or remove columns.
    """

    # different types (auto-detect) for this time series
    TYPE_DICT_OF_LISTS = "dict_list"
    TYPE_LIST_OF_DICTS = "list_dict"

    # initial resolution (auto-detect)
    resolution = None

    def __init__(self, input_data, date_time_key="date_time", date_format=None,
                 time_zone=None):
        """
        Creates a new time series instance.
        :param input_data: the input data, which should be either a list of dictionaries, containing at least a date
        time key with values OR a map of lists, where the map contains at least a date time key with a list of values
        underneath.
        :param date_time_key: the key under which to find the date time information
        :param date_format: optional date format to parse dates if they're provided as string.
        :param time_zone: optional time zone for the date time column. Will default to the current time zone of the
        machine the code is running on.
        """
        # fallback for timezone internally, because Python older than 3.6 causes issues when importing pbu
        self.time_zone = time_zone
        if self.time_zone is None:
            self.time_zone = datetime.utcnow().astimezone().tzinfo

        # store fields
        self.data = input_data
        self.date_time_key = date_time_key
        self.date_format = date_format
        self.time_zone = time_zone

        # analyse input data
        self.type, self.keys = self._check_input_data()
        # parse dates if necessary
        self._parse_dates()
        # extract resolution if possible as well
        if self.type is not None:
            self.set_resolution()

    def set_resolution(self, custom_resolution=None):
        """
        Sets the resolution for the time series. If not resolution is provided, it will be detected automatically. The
        method doesn't return anything, but rather updates the resolution attribute of this instance.
        :param custom_resolution: a timedelta instance specifiying the gap between 2 date time values.
        """
        if custom_resolution is None:
            self.resolution = self.get_resolution()
        else:
            self.resolution = custom_resolution

    def get_resolution(self):
        """
        Auto-detects the resolution of the current date time column series. The method will iterate through all values
        of the date time column and measure the difference between timestamps, create a statistic for each difference
        and returns the most prominent result in the time series.
        :return: the most common timedelta between date time column values in the data.
        """
        last_dt = datetime.now().astimezone(self.time_zone)
        # collect stats
        stats = {}
        # determine datetime series from input data
        dt_series = self.get_dates()
        for dt in dt_series:
            if isinstance(dt, str):
                # string dates, need to be parsed
                if self.date_format is None:
                    raise AttributeError("Attempted to determine resolution, but didn't provide a date format for "
                                         "the date time column values, which are strings like: {}".format(dt))
                dt = datetime.strptime(dt, self.date_format).astimezone(self.time_zone)

            # check the diff from last datetime and collect diff in stats
            diff = dt - last_dt
            if diff not in stats:
                stats[diff] = 0
            stats[diff] += 1
            last_dt = dt

        # evaluate collected stats and find most common resolution
        res_max = 0
        current_max = None
        for resolution in stats:
            if stats[resolution] > res_max:
                res_max = stats[resolution]
                current_max = resolution

        if current_max is None:
            raise ValueError("Resolution could not be determined. This is most likely due to missing data.")
        return current_max

    def add_values(self, new_key, new_values):
        """
        Adds a new series to the existing one. We assume that the new series is already aligned to the date time column
        of this time series. The new values will simply be added to the existing data.
        :param new_key: the key under which the new data series will be available
        :param new_values: a list of all the values for the this key over the whole data series.
        """
        # don't add duplicate keys
        if new_key in self.keys:
            raise AttributeError("Can't add key '{}' to given time series, because it already contains that "
                                 "key".format(new_key))

        # check length of new data
        existing_data_length = len(self.data)
        if self.type == TimeSeries.TYPE_DICT_OF_LISTS:
            existing_data_length = len(self.data[self.date_time_key])

        # make sure the value series align in length, if not fill new series with values (max +2)
        new_values = TimeSeries._align_value_series_length(new_values, existing_data_length)
        if len(new_values) != existing_data_length:
            # difference in size is more than 2
            raise ValueError(
                "Failed to add new data series with length {} to existing data series with length {}".format(
                    len(new_values), existing_data_length))

        # add data
        if self.type == TimeSeries.TYPE_DICT_OF_LISTS:
            self.data[new_key] = new_values

        elif self.type == TimeSeries.TYPE_LIST_OF_DICTS:
            # translate to dictionary, add values and then translate back
            output = self.translate_to_dict_of_lists()
            output[new_key] = new_values
            self.data = TimeSeries(output, self.date_time_key, time_zone=self.time_zone).translate_to_list_of_dicts()

        self.keys.append(new_key)

    def fill_values(self, new_key, constant_value):
        """
        Adds a new data series to the existing data by filling it up completely with a constant value. The method does
        not return any results, but instead updates the instance's data directly.
        :param new_key: the key under which to store the new series
        :param constant_value: the constant value to add for each entry in the new series
        """
        # check for duplicates
        if new_key in self.keys:
            raise AttributeError(
                "Can't add key '{}' to given time series as constant, because it already exists.".format(new_key))

        # add data
        if self.type == TimeSeries.TYPE_DICT_OF_LISTS:
            self.data[new_key] = [constant_value] * len(self.data[self.date_time_key])
        else:
            output = self.translate_to_dict_of_lists()
            output[new_key] = [constant_value] * len(self.data)
            self.data = TimeSeries(output, self.date_time_key, time_zone=self.time_zone).translate_to_list_of_dicts()

    def get_values(self, selected_key):
        """
        Extracts the values of a given column as list from the current time series.
        :param selected_key: the key to retrieve
        :return: a list of values representing this series
        """
        if self.type == TimeSeries.TYPE_DICT_OF_LISTS:
            # check if the key is available
            if selected_key not in self.data:
                raise ValueError("Requested key {} could not be found in input data".format(selected_key))
            # return the series
            dt_series = self.data[selected_key]
            return dt_series
        elif self.type == TimeSeries.TYPE_LIST_OF_DICTS:
            # check if the key exists in the first item
            if selected_key not in self.data[0]:
                raise ValueError("Requested key {} could not be found in first item of input data".format(selected_key))
            # return the extracted column
            dt_series = list(map(lambda x: x[selected_key], self.data))
            return dt_series

        raise AttributeError("Data series doesn't have a valid type. Extraction of value series not possible.")

    def get_dates(self):
        """
        Extracts the values of the date time column as list from the current time series.
        :return: a list of datetime objects
        """
        if self.type == TimeSeries.TYPE_DICT_OF_LISTS:
            return self.data[self.date_time_key]
        elif self.type == TimeSeries.TYPE_LIST_OF_DICTS:
            return list(map(lambda x: x[self.date_time_key], self.data))

        raise AttributeError("Data series doesn't have a valid type. Extraction of date time series not possible.")

    def get_start_date(self):
        """
        Extracts the first date in the date time series.
        :return: a datetime object representing the first entry in the time series.
        """
        return self._get_date_value(0)

    def get_end_date(self):
        """
        Extracts the last date in the date time series.
        :return: a datetime object representing the last entry in the time series
        """
        return self._get_date_value(-1)

    def add_series(self, time_series, keys_to_add=None):
        """
        Adds a new time series to this existing one. The first step is to align the new time series with this instance's
        date time. Then the columns of the argument will be added one by one to this instance. The method doesn't return
        anything, but directly updates the existing data.
        :param time_series: a time series instance to add
        :param keys_to_add: optional a list of keys to add that exist in the new series. If omitted, the method will add
        all columns it finds (except the date time column).
        """
        # check type
        if not isinstance(time_series, TimeSeries):
            raise ValueError("Provided time series is not of type TimeSeries, but {}".format(type(time_series)))
        # prepare keys
        if keys_to_add is None:
            # fallback to all keys of the series
            keys_to_add = time_series.keys

        # check for duplicate keys
        for key in keys_to_add:
            if key in self.keys:
                raise ValueError("Attempting to add key {} to existing data series, which already has "
                                 "such a key".format(key))

        # first align current data set (ensure it is aligned)
        resolution = self.get_resolution()  # resolution of existing series
        self.align_to_resolution(resolution)

        # also align the new time series to the given resolution and time frame
        time_series.align_to_resolution(resolution=resolution, start_date=self.get_start_date(),
                                        end_date=self.get_end_date())
        for key in time_series.keys:
            self.add_values(key, time_series.get_values(key))
        _, self.keys = self._check_input_data()

    def remove_series(self, keys_to_remove):
        """
        Removes a column from the time series. The method will not return anything, but directly modify the data of this
        instance.
        :param keys_to_remove: the key to remove
        """
        if isinstance(keys_to_remove, str):
            keys_to_remove = [keys_to_remove]

        if self.date_time_key in keys_to_remove:
            raise AttributeError("Can't remove date time key {}".format(self.date_time_key))

        if self.type == TimeSeries.TYPE_DICT_OF_LISTS:
            for key in keys_to_remove:
                del self.data[key]
        else:
            # translate
            output = self.translate_to_dict_of_lists()
            for key in keys_to_remove:
                del output[key]
            self.data = TimeSeries(output, self.date_time_key, time_zone=self.time_zone).translate_to_list_of_dicts()
        _, self.keys = self._check_input_data()

    def align_to_resolution(self, resolution=None, start_date=None, end_date=None):
        """
        Aligns the current input data to the most dominant resolution. This will keep the first date of the datetime
        column and then interpolate values in between and use existing values to fill the data series with a fixed
        interval between each data point. This will update the original data of this time series. The method doesn't
        return anything, but directly modifies the instance's data.
        :param resolution: optional, if provided will force a specific resolution and interpolate/skip values to match
        that resolution
        :param start_date: an optional start date, in case you want to prepend or cut the existing data series to a
        specific time frame. If not provided the first date from the time series will be used.
        :param end_date: an optional end date, in case you want to append or cut the existing data series to a specific
        time frame. If not provided, the last date from the time series will be used.
        """
        # fetch resolution and tolerance
        if resolution is None:
            resolution = self.get_resolution()
        tolerance = resolution / 2

        # start date (use start date minus resolution to match the first entry perfectly)
        current_date = self.get_start_date()
        if start_date is not None:
            current_date = start_date
        if end_date is None:
            end_date = self.get_end_date()

        # standardise initial data
        main_data = self.data
        if self.type == TimeSeries.TYPE_DICT_OF_LISTS:
            main_data = self.translate_to_list_of_dicts()

        # runtime variables for the alignment operation below
        prev_value = None
        current_original_data_index = 0
        result = []
        # as long as we haven't reached the end date, keep adding values
        while current_date <= end_date:
            is_last = False
            if current_original_data_index >= len(main_data) - 1:
                is_last = True

            # ensure we don't run into index errors (repeat last element until end-date in worst-case scenario)
            if current_original_data_index >= len(main_data):
                current_original_data_index = len(main_data) - 1  # might cause endless loop in skip branch

            # extract date information and time delta
            original_date = main_data[current_original_data_index][self.date_time_key]
            diff_seconds = (current_date - original_date).total_seconds()

            # close enough
            if current_date == original_date or abs(diff_seconds) < tolerance.total_seconds():
                # fetch the value from the original data
                prev_value = main_data[current_original_data_index]
                # update date of this entry to current date
                prev_value[self.date_time_key] = current_date
                # add to result
                result.append(prev_value)
                # remember to increment the key for next item and add resolution to current date
                current_original_data_index += 1
                current_date += resolution

            # implies the data point is no where close, skip this entry
            elif current_date > original_date:
                prev_value = main_data[current_original_data_index]

                if is_last:
                    # we've already reached the end of the data series, fill with last value
                    if current_date < end_date:
                        missing_entries = math.floor((end_date - current_date) / resolution) + 1
                        for index in range(0, missing_entries):
                            # copy last item
                            new_item = copy(prev_value)
                            # update timestamp
                            new_item[self.date_time_key] = current_date + resolution
                            # add to result
                            result.append(new_item)
                            # remember for next loop
                            prev_value = new_item
                            # increment timestamp
                            current_date += resolution
                    if current_date == end_date:
                        # breaks the while loop

                        current_date += resolution
                else:
                    current_original_data_index += 1

            # implies that the next data point is in the future and we need to interpolate missing values
            elif current_date < original_date:
                # compute gaps
                gaps = {}
                next_value = main_data[current_original_data_index]

                if prev_value is None:
                    prev_value = next_value
                    # special handling for initial result to prefill with first value
                    if len(result) == 0:
                        prev_value[self.date_time_key] = current_date
                for key in self.keys:
                    gaps[key] = next_value[key] - prev_value[key]
                missing_entries = math.floor((original_date - current_date) / resolution)
                # interpolate entries starting from prev_value
                for index in range(0, missing_entries):
                    new_item = {
                        self.date_time_key: prev_value[self.date_time_key] + resolution
                    }
                    for key in self.keys:
                        new_item[key] = prev_value[key] + (gaps[key] / missing_entries)

                    prev_value = new_item
                    result.append(new_item)
                    # update current date with latest added value
                    current_date = new_item[self.date_time_key]
                # increment resolution and index
                current_date += resolution
                current_original_data_index += 1

        # store aligned data
        if self.type == TimeSeries.TYPE_DICT_OF_LISTS:
            # convert if necessary
            self.data = TimeSeries(input_data=result,
                                   date_time_key=self.date_time_key,
                                   time_zone=self.time_zone).translate_to_dict_of_lists()
        else:
            # already in correct format
            self.data = result

    def translate_to_list_of_dicts(self, date_format=None):
        """
        Returns the data of this instance as a list containing dictionary items with all keys and the datetime key
        being contained in each item of the list.
        :param date_format: optional date format to render date_time columns as strings rather than native datetime
        :return: a list of dictionaries
        """
        if self.type == TimeSeries.TYPE_LIST_OF_DICTS:
            # already in correct format
            result = copy(self.data)
            # format date_time column if necessary
            if date_format is not None:
                result = list(map(lambda x: TimeSeries._format_date_list_of_dict(x, self.date_time_key, date_format),
                                  result))
            return result
        else:
            # need translation
            result = []
            # iterate through all values
            for index in range(0, len(self.data[self.date_time_key]), 1):
                # iterate through all keys for current index and add to result
                item = {}
                for key in self.data:
                    value = self.data[key][index]
                    if key == self.date_time_key and date_format is not None:
                        # handle datetime rendering
                        value = value.strftime(date_format)
                    # add key to current item
                    item[key] = value
                # append finished item to list of dicts
                result.append(item)
            # return list of dicts
            return result

    def translate_to_dict_of_lists(self, date_format=None):
        """
        Returns the data of this instance as a dictionary with keys for each data series, including the date time key.
        Underneath each key, all values (sorted by date time) are stored in a simple list containing the values as
        primitive types.
        :param date_format: optional date format to render date_time columns as strings rather than native datetime
        objects
        :return: a dictionary containing lists
        """
        if self.type == TimeSeries.TYPE_DICT_OF_LISTS:
            # already in correct format
            result = copy(self.data)
            # format date_time column if necessary
            if date_format is not None:
                result[self.date_time_key] = list(map(lambda x: x.strftime(date_format), result[self.date_time_key]))
            return result
        else:
            # need translation
            result = {}
            # init keys
            for key in self.data[0]:
                result[key] = []

            # map to lists
            for item in self.data:
                for key in item:
                    value = item[key]
                    if key == self.date_time_key and date_format is not None:
                        # handle datetime rendering
                        value = value.strftime(date_format)
                    result[key].append(value)
            # return dict of lists
            return result

    def to_pd_data_frame(self):
        """
        Converts the time series into a pandas DataFrame with a given time index
        :return: a pandas DataFrame with the date_time column set as the index
        """
        df = pd.DataFrame(self.translate_to_dict_of_lists(date_format=None)).set_index(self.date_time_key)
        for col in df:
            if col != self.date_time_key:
                df[col] = pd.to_numeric(df[col], errors='coerce')
        return df

    @staticmethod
    def create_date_range(from_date, to_date=None, num_points=None, resolution=timedelta(minutes=5),
                          include_start_date=False, time_zone=None):
        """
        Creates a list of datetime instances in a given resolution an a given time frame.
        :param from_date: the start date for the time series
        :param to_date: the end date for the time series (default: None)
        :param num_points: optional: the number of points to add (use only num_points or to_date)
        :param resolution: the resolution for the time series (default 5 minutes)
        :param include_start_date: boolean flag to indicate whether to add the from date or not (default: False)
        :param time_zone: optional the timezone for which to create the date range (default UTC)
        :return: a list of datetime objects matching the provided resolution and from/to date.
        """
        if num_points is not None and to_date is not None:
            raise AttributeError("Cannot provide to_date and num_points at the same time")

        # internal fallback to default, to avoid import issues with old python versions
        effective_time_zone = time_zone
        if effective_time_zone is None:
            effective_time_zone = datetime.utcnow().astimezone().tzinfo

        # localize dates
        from_date = from_date.astimezone(effective_time_zone)
        if to_date is not None:
            to_date = to_date.astimezone(effective_time_zone)

        # optionally offset start date
        if not include_start_date:
            from_date += resolution

        # if num_points are provided, compute end date
        if num_points is not None:
            num_points -= 1
            to_date = from_date + (num_points * resolution)

        if from_date > to_date:
            return [from_date]

        result = []
        current_date = from_date - resolution
        while current_date < to_date:
            current_date += resolution
            result.append(current_date)

        return result

    @staticmethod
    def _format_date_list_of_dict(item, date_time_key, date_format):
        item[date_time_key] = item[date_time_key].strftime(date_format)
        return item

    @staticmethod
    def _align_value_series_length(new_values, length):
        if len(new_values) + 1 == length:
            # missing one value, append last value again
            new_values.append(new_values[-1])
        elif len(new_values) + 2 == length:
            # missing two values, prepend first, append last
            new_values = [new_values[0]] + new_values + [new_values[-1]]
        elif len(new_values) - 1 == length:
            # remove last item
            new_values = new_values[0:len(new_values) - 1]
        elif len(new_values) - 2 == length:
            # remove first and last
            new_values = new_values[1:len(new_values) - 1]
        return new_values

    def _get_date_value(self, index):
        """
        Extracts a given date from the date time series.
        :param index: the index to retrieve (can be any valid Python list accessor)
        :return: the date time at the given index
        :raise IndexError, if the index provided doesn't exist.
        """
        if self.date_time_key is None:
            raise AttributeError("Current time series doesn't provide a date_time_key")

        if self.type == TimeSeries.TYPE_LIST_OF_DICTS:
            return self.data[index][self.date_time_key]
        elif self.type == TimeSeries.TYPE_DICT_OF_LISTS:
            return self.data[self.date_time_key][index]

    def _check_input_data(self):
        """
        Helper method to determine the type of the input data and the extract the keys.
        :return: a tuple containing (1) the type of the input data and (2) the keys contained in the input data.
        """
        if self.data is None or self.date_time_key is None:
            # no association between input data and date_time
            return None, []

        if isinstance(self.data, (dict, JSON)) and self.date_time_key in self.data:
            # dictionary of data series, including other keys
            result = []
            for key in self.data:
                if key != self.date_time_key:
                    result.append(key)
            return TimeSeries.TYPE_DICT_OF_LISTS, result

        if isinstance(self.data, list) and len(self.data) > 0 and isinstance(self.data[0], (dict, JSON)):
            # list of dictionaries
            result = []
            for key in self.data[0]:
                if key != self.date_time_key:
                    result.append(key)
            return TimeSeries.TYPE_LIST_OF_DICTS, result

        if isinstance(self.data, list) and len(self.data) == 0:
            return TimeSeries.TYPE_LIST_OF_DICTS, []

        raise ValueError("Provided input data structure is not supported. Only a dictionary of lists or a list of "
                         "dictionaries is allowed. Found type of input_data: {}".format(type(self.data)))

    def _parse_dates(self):
        """
        Helper method to parse dates in the date column. If the dates are already dates, no change is performed. It
        can detect unix timestamps (float or int) and string representation of dates (using the date_format provided in
        the constructor).
        The method does not return any result, but rather modifies the data stored in this instance.
        """
        if len(self.get_dates()) == 0:
            if self.type == TimeSeries.TYPE_DICT_OF_LISTS:
                self.data[self.date_time_key] = []
            return
        dates = self.get_dates()
        if isinstance(dates[0], datetime):
            # already in correct format
            return

        converted_dates = []
        if isinstance(dates[0], (float, int)):
            # timestamp
            for dt in dates:
                converted_dates.append(datetime.fromtimestamp(dt, tz=self.time_zone))
        elif isinstance(dates[0], str):
            if self.date_format is None:
                raise AttributeError("Could not parse dates, because of missing date format.")
            # datetime string
            for dt in dates:
                converted_dates.append(datetime.strptime(dt, self.date_format).astimezone(self.time_zone))
        else:
            raise ValueError("Could not parse dates of input data set for type {}".format(type(dates[0])))

        # update existing dates
        if self.type == TimeSeries.TYPE_DICT_OF_LISTS:
            self.data[self.date_time_key] = converted_dates
        else:
            # translate
            output = self.translate_to_dict_of_lists()
            output[self.date_time_key] = converted_dates
            self.data = TimeSeries(output, date_time_key=self.date_time_key,
                                   time_zone=self.time_zone).translate_to_list_of_dicts()



import os
import json
from typing import Optional, Union


def write_json(data: Union[dict, list], path: str):
    """
    Writes an object to a json file. This function will take care of opening and closing the file.
    :param data: a list of dictionary - has to be possible to serialise it as JSON.
    :param path: the file path where to write the file to
    """
    if data is None or not isinstance(data, (list, dict)):
        raise ValueError("No or invalid data provided")
    fp = open(path, "w")
    json.dump(data, fp)
    fp.close()


def read_json(path: str) -> Optional[Union[dict, list]]:
    """
    Reads from a json file if it exists and returns the content. This function will take care of opening and closing the
    file.
    :param path: the file path of the json file
    :return: the list of dictionary contained in the json file
    """
    if not os.path.exists(path):
        return None
    fp = open(path)
    data = json.load(fp)
    fp.close()
    return data



from typing import List, Union, Optional
from statistics import mean


def weighted_mean(values: List[Union[float, int]], weights: List[Union[int, float]] = []) -> Optional[float]:
    """
    This will generate a mean value for the list of provided `values`, where each value is multiplied by the
    corresponding weight in the same position. If there are more `values` than `weights`, remaining values will receive
    weight 1.0. If there's more `weights` than `values`, it will only use the first `n` weights, where `n` is the number
    of values.
    :param values: the numeric values to generated a weighted mean over
    :param weights: the weights for each position of the values
    :return: the weighted average value of the provided values
    """
    if len(values) == 0:
        return None

    if weights is None or len(weights) == 0:
        return mean(values)

    total_weight = 0.0
    total_value = 0.0

    num_weights = len(weights)

    for idx, val in enumerate(values):
        current_weight = weights[-1] if num_weights < idx + 1 else weights[idx]
        total_weight += current_weight
        total_value += val * current_weight

    if total_weight == 0.0:
        return None

    return total_value / total_weight


def normalise(value: Union[float, int], min_val: Union[float, int], max_val: Union[float, int], limit=True,
              mid_point: Optional[Union[float, int]] = None) -> float:
    """
    Normalises the given input `value` between min_val and max_val as number between 0 and 1. If the min_val is provided
    larger than the max_val, the function will automatically invert them and provide the `1 - normalised_value`.
    :param value: the number to normalise
    :param min_val: the lower boundary representing 0.0
    :param max_val: the upper boundary representing 1.0
    :param limit: a boolean flag indicating whether a value between 0 and 1 is returned. Otherwise, values greater than
    1 and lower than 0 can be returned.
    :param mid_point: an optional mid-point to use for the normalisation. If provided, the normalisation will be
    performed between the min_val and the mid_point, and the mid_point and the max_val.
    :return: the normalised value expressed as where the value sits between min_val and max_val
    """
    if value is None or min_val is None or max_val is None:
        return 0.0  # invalid input

    min_value = min_val
    max_value = max_val
    inverted = False
    if max_val < min_val:
        inverted = True
        min_value = max_val
        max_value = min_val

    if value < min_value and limit is True:
        return 0.0 if inverted is False else 1.0

    if value > max_value and limit is True:
        return 1.0 if inverted is False else 0.0

    if max_value == min_value:
        return 0.5

    norm = 0.5  # just in case
    if mid_point is None:
        norm = (float(value) - float(min_value)) / (float(max_value) - float(min_value))
    else:
        if value <= mid_point:
            norm = ((float(value) - float(min_value)) / (float(mid_point) - float(min_value))) * 0.5
        else:
            norm = (((float(value) - float(mid_point)) / (float(max_value) - float(mid_point))) * 0.5) + 0.5
    return norm if inverted is False else 1.0 - norm


def discretise(value: Union[float, int], precision: Union[float, int] = 1, floor=False,
               ceil=False) -> Union[float, int]:
    """
    This function will round the given value to the nearest multiple of the given precision. There are 2 boolean flags
    available that can force the function to return the lower or upper boundary of a precision interval. If neither is
    set, the function will return the mid point of the interval.
    :param value: the value to discretise
    :param precision: the precision to use for the discretisation
    :param floor: a boolean flag indicating whether the lower boundary of the precision interval should be returned
    :param ceil: a boolean flag indicating whether the upper boundary of the precision interval should be returned
    :return: the discretised value
    """
    if value is None:
        return None
    remainder = value % precision
    if ceil is True:
        return value - remainder + precision
    if floor is True:
        return value - remainder
    # mid point is default fallback
    return (value - remainder) + (precision * 0.5)



from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(name="pbu",
      version="1.0.3",
      description="Basic Utility module for the Python programming language",
      long_description=long_description,
      long_description_content_type="text/markdown",
      url="https://github.com/ilfrich/python-basic-utils",
      author="Peter Ilfrich",
      author_email="das-peter@gmx.de",
      license="Apache-2.0",
      packages=[
          "pbu"
      ],
      install_requires=[
          "pytz",
          "requests",
          "pandas",
          "python-dotenv",
      ],
      tests_require=[
          "pytest",
      ],
      zip_safe=False)



import warnings
from abc import ABC, abstractmethod
from typing import Optional, List


class JsonDocument(ABC):
    def __init__(self):
        pass

    def extract_system_fields(self, json: dict):
        """
        Extracts the id and version from a JSON object or dictionary and maps them to the current instances attributes.
        :param json: the json object or dictionary from which to extract information.
        """
        # check if the get_attribute_mapping method is overridden
        attr_mapping = self._get_attribute_mapping()
        if attr_mapping is None:
            return

        # evaluate attribute mapping
        for key in attr_mapping:
            if attr_mapping[key] in json:
                self.__setattr__(key, json[attr_mapping[key]])

    def to_json(self) -> dict:
        """
        Returns a serializable representation of this document as dictionary or JSON object.
        :return: a dictionary or JSON object providing the data contained within this document
        """
        result = {}
        attr_mapping = self._get_attribute_mapping()
        if attr_mapping is not None:
            for key in attr_mapping:
                if self.__getattribute__(key) is not None:
                    result[attr_mapping[key]] = self.__getattribute__(key)

        return result

    @staticmethod
    @abstractmethod
    def from_json(json: dict):
        """
        Receives a dictionary or JSON object and returns an instance of this JsonDocument subclass.
        :param json: a dictionary or JSON object instance
        :return: an instance of a subclass of JsonDocument
        """
        pass

    def get_attribute_mapping(self) -> dict:
        """
        Provides a mapping from internal attribute names to JSON attribute names.
        """
        pass

    def _get_attribute_mapping(self) -> Optional[dict]:
        """
        Internal method used to find out if the subclass defines an attribute mapping. If the subclass defines an
        attribute mapping and returns a dictionary, the attribute mapping will be returned. Otherwise, None will be
        returned, which will be used by the to_json and extract_system_fields method to map all primitive fields from
        the de-serialised class to JSON and back.
        """
        # find out if the subclass defines the method
        defining_class = self.get_attribute_mapping.__func__.__qualname__.split(".")[0]
        if defining_class == "JsonDocument":
            return None

        # check if the subclass method returns a dictionary
        attr_mapping = self.get_attribute_mapping()
        if not isinstance(attr_mapping, dict):
            return None

        # mapping provided by subclass, return it
        return attr_mapping

    def apply_updates(self, update, attributes: List[str] = []):
        """
        Applies an update (which has to be of the same type as self) to the current instance. The list of attributes
        past will be checked, if they are available.
        :param update: an instance of the same type as self
        :param attributes: a list of strings representing the attributes to update (see get_attribute_mapping)
        :raises ValueError: in case the provided `update` is an instance of a different class than self.
        """
        if not isinstance(update, type(self)):
            raise ValueError("Provided `update` parameter has a different type than `self`")
        if update is None:
            return
        for attr in attributes:
            if hasattr(self, attr) and hasattr(update, attr):
                self.__setattr__(attr, getattr(update, attr))
            else:
                warnings.warn(f"Trying to update attribute '{attr}' but either the item or the update does not have it")


def list_to_json(item_list: List[JsonDocument]):
    """
    Helper class serialising a list of JsonDocument into a list of dictionaries that can easily be serialised.
    :param item_list: a list of JsonDocument instances
    :return: a list of dictionaries.
    """
    return list(map(lambda item: item.to_json(), item_list))


def list_from_json(json_list: List[dict], deserialize_class: JsonDocument):
    """
    Helper class deserialising a list of dictionaries into a list of JsonDocument instances.
    :param json_list: a list of dictionaries
    :param deserialize_class: the class to use for deserialisation
    :return: a list of JsonDocument instances
    """
    if not issubclass(deserialize_class, JsonDocument):
        raise ValueError("Provided `deserialize_class` is not a subclass of JsonDocument")
    return list(map(lambda json: deserialize_class.from_json(json), json_list))



from pbu.json_wrapper import JSON
from pbu.logger import Logger
from pbu.time_series import TimeSeries
from pbu.paging import PagingInformation
from pbu.basic_monitor import BasicMonitor, JobStatus
from pbu.default_options import default_options, default_value, list_find_one, list_map_filter, list_join, not_none
from pbu.constant_listing import ConstantListing
from pbu.performance_logger import PerformanceLogger
from pbu.date_time import combine_date_time, to_utc, to_timezone, set_timezone, DATE_FORMAT, DATETIME_FORMAT
from pbu.datascience_util import weighted_mean, normalise, discretise
from pbu.app_config import BasicConfig
from pbu.json_document import JsonDocument, list_to_json, list_from_json
from pbu.files import write_json, read_json



import os
import requests
import logging
import sys
import traceback
from logging import handlers

CONFIG_KEY_LOG_SERVER = "PBU_LOG_SERVER"
CONFIG_KEY_LOG_SERVER_AUTH = "PBU_LOG_SERVER_AUTH"
CONFIG_KEY_LOG_FOLDER = "PBU_LOG_FOLDER"


class _CustomHttpHandler(logging.Handler):
    def __init__(self, level, url, auth_token=None):
        logging.Handler.__init__(self, level)
        self.url = "{}/api/log".format(url)
        self.auth_token = auth_token

    @staticmethod
    def _map_log_record(record):
        """
        Default implementation of mapping the log record into a dict
        that is sent as the CGI data. Overwrite in your class.
        Contributed by Franz Glasner.
        """
        result = record.__dict__
        if "exc_info" in result and result["exc_info"] is not None:
            trace = []
            for el in result["exc_info"]:
                if type(el).__name__ == "traceback":
                    # custom traceback logging, because we can't serialise this
                    stack = traceback.extract_stack()
                    for err_line in stack[:-9]:
                        trace.append("    {}:{} {}".format(err_line.filename, err_line.lineno, err_line.name))
                        trace.append("        {}".format(err_line.line))
                    for err_line in traceback.format_tb(el):
                        trace.append("    {}".format(err_line.strip()))
            del result["exc_info"]
            if len(trace) > 0:
                result["trace"] = trace
        if "msg" in result and not isinstance(result["msg"], str):
            result["msg"] = str(result["msg"])
        return result

    def emit(self, record):

        is_higher_than_error = self.level == logging.ERROR and record.levelno in [logging.CRITICAL, logging.FATAL]
        if record.levelno != self.level and not is_higher_than_error:
            return

        headers = {
            "Content-Type": "application/json",
        }
        if self.auth_token is not None:
            headers["Authorization"] = self.auth_token
        try:
            requests.post(url=self.url, json=_CustomHttpHandler._map_log_record(record), headers=headers)
        except BaseException as be:
            print("Error sending log message: {} ({})".format(_CustomHttpHandler._map_log_record(record), be),
                  file=sys.stderr)


class Logger(logging.Logger):
    """
    File logger for this application, logging into application.log in the configured LOG_FOLDER.
    Usage:

    >>> logger = Logger("some-name")
    >>> logger.info("My message")
    """

    def __init__(self, name, log_folder=None, enable_logger_name=True, enabled_log_levels=[logging.INFO, logging.ERROR],
                 message_format="%(asctime)s %(levelname)s:%(name)s %(message)s"):
        """
        Creates a new instance of this logger and will store it as a private field, which is exposed via the get()
        method.
        :param name: the name of the class / component, which will be added as a marker to each log.
        """
        name = name.replace(".log", "")
        logger = logging.getLogger(name)
        logger.setLevel(logging.DEBUG)

        # decide if this logger sends messages to a log server
        self.log_server = os.getenv(CONFIG_KEY_LOG_SERVER)
        self.log_server_auth = os.getenv(CONFIG_KEY_LOG_SERVER_AUTH)

        self.is_worker = self.log_server is not None

        if log_folder is not None:
            if not os.path.isdir(log_folder):
                os.makedirs(log_folder)

        # check if other handlers are provided
        if not logger.handlers:
            if self.is_worker:
                # worker process
                self._configure_worker(logger, self.log_server, message_format, self.log_server_auth,
                                       enabled_log_levels)
            else:
                # listener process
                if os.getenv(CONFIG_KEY_LOG_FOLDER) is not None and log_folder is None:
                    log_folder = os.getenv(CONFIG_KEY_LOG_FOLDER)
                if log_folder is None:
                    self._configure_listener(logger, message_format, enable_logger_name=enable_logger_name,
                                             enabled_log_levels=enabled_log_levels)
                else:
                    self._configure_listener(logger, message_format, log_folder=log_folder,
                                             enable_logger_name=enable_logger_name,
                                             enabled_log_levels=enabled_log_levels)

        self._logger = logger

    def warn(self, msg, *args, **kwargs):
        self._logger.warning(msg, *args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        self._logger.warning(msg, *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        self._logger.error(msg, stack_info=True, exc_info=True, *args, **kwargs)

    def debug(self, msg, *args, **kwargs):
        self._logger.debug(msg, *args, **kwargs)

    def info(self, msg, *args, **kwargs):
        self._logger.info(msg, *args, **kwargs)

    def exception(self, msg):
        self._logger.exception(msg, stack_info=True)

    def handle(self, record):
        self._logger.handle(record)

    def get_handler(self):
        if len(self._logger.handlers) == 0:
            return None
        return self._logger.handlers[0]

    def __repr__(self):
        return self._logger.__repr__()

    @staticmethod
    def _configure_worker(logger, url, message_format, auth=None, enabled_log_levels=[logging.INFO, logging.ERROR]):
        for log_level in enabled_log_levels:
            handler = _CustomHttpHandler(log_level, url, auth)
            formatter = logging.Formatter(message_format)
            handler.setFormatter(formatter)
            logger.addHandler(handler)

    @staticmethod
    def _configure_listener(logger, message_format, log_folder="_logs", enable_logger_name=True,
                            enabled_log_levels=[logging.INFO, logging.ERROR]):
        formatter = logging.Formatter(message_format)
        if enable_logger_name is False:
            # remove logger name from message format
            message_format = message_format.replace("%(name)s", "")
            formatter = logging.Formatter(message_format)

        file_names = {
            logging.INFO: "info.log",
            logging.DEBUG: "debug.log",
            logging.ERROR: "error.log",
            logging.WARNING: "warning.log",
        }

        for log_level in enabled_log_levels:
            file_name = os.path.join(log_folder, file_names[log_level])
            if not os.path.isdir(log_folder):
                os.makedirs(log_folder)
            handler = handlers.TimedRotatingFileHandler(file_name, when="d", interval=1, backupCount=30)
            handler.setFormatter(formatter)
            handler.setLevel(log_level)
            logger.addHandler(handler)



import os
from typing import List, Dict, Any, Optional
from dotenv import load_dotenv


class BasicConfig:
    def __init__(self, default_values: Dict[str, Any] = {}, directory_keys: List[str] = [], required: List[str] = [],
                 env_file=".env"):
        self.config = {}
        self.default_values = default_values
        self.directory_keys = directory_keys
        self.required_keys = required
        self._env_file = env_file
        self._load_config()

    def _load_config(self):
        # read out existing os environment
        load_dotenv(self._env_file)
        self.config = {}

        # apply defaults for missing config params
        for key in self.default_values:
            val = os.getenv(key)
            if key in self.required_keys and val is None:
                raise EnvironmentError(f"You need to provide an environment variable specifying {key}")
            self.config[key] = self.default_values[key] if val is None else val

        # check that all directories exist
        for dir_type in self.directory_keys:
            self._create_directory(dir_type)

    def get_config_value(self, config_key: str, default_value: Optional[Any] = None) -> Optional[Any]:
        if config_key in self.config:
            return self.config.get(config_key, default_value)

        # read fresh from environment
        val = os.getenv(config_key)
        # store in local cache and return
        self.config[config_key] = val
        return val

    def _create_directory(self, config_key=str):
        if self.config.get(config_key, None) is None:
            return
        current_dir = self.config[config_key]
        if not os.path.isdir(current_dir):
            os.mkdir(current_dir)
